/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2005, 2006, Digium, Inc.
 *
 * redice li <redice_li@yahoo.com>
 * Russell Bryant <russell@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 * 
 * \brief A failover module for openvox failover products
 * Copyright (C) 2008 OpenVox, Inc.
 * \author <mark.liu@openvox.cn>
 */ 


#include "asterisk.h"
#include <termios.h>
#include <errno.h>
#include <sysexits.h>
#include <ctype.h>

#include "asterisk/channel.h"
#include "asterisk/logger.h"
#include "asterisk/module.h"
#include "asterisk/cli.h"

#ifndef AST_MODULE_INFO
#define AST_VERSION_1_2
#elif !defined AST_CLI_DEFINE
#define AST_VERSION_1_4
#elif defined AST_CLI_DEFINE
#define AST_VERSION_1_6
#endif 

#ifdef AST_VERSION_1_2
#define __CONST__
#define __static__
#include <stdio.h>
#include <stdlib.h>
#else
#define __CONST__ const
#define __static__ static
#endif
	
#include "asterisk/file.h"


#define MAX_DEVICE_NAME 128
#define MAX_CMD_NAME 64
#define MAX_PARAM_NAME 8

/* serial port default value */
#define DEFAULT_SPEED	"1200"
#define DEFAULT_PARMS	"8n1"

/* failover config file */
#define FAILOVER_CONFIG_FILE "failover.conf"

/* failover serial device */
typedef struct tty_struct {
    int             fd;
    char            tty_name[MAX_DEVICE_NAME];
    struct termios  termios_state;
    speed_t         baudrate;
}tty_struct_t;
static int failover_dev_num;

/* event */
#define FAILOVER_EVENT_STARTING		"STARTING"
#define FAILOVER_EVENT_RUNNING		"RUNNING"
#define FAILOVER_EVENT_WDOG_OK		"WDOG_OK"
#define FAILOVER_EVENT_WDOG_FAILED	"WDOG_FAILED"
#define FAILOVER_EVENT_STOPPED		"STOPPED"
#define FAILOVER_EVENT_ERROR		"ERROR"

/* failover device */
typedef struct failover_dev {
	tty_struct_t tty;
	char dev_name[MAX_DEVICE_NAME];
	int kick_time_interval;
	char event_cmd[MAX_CMD_NAME];
	char autorun[MAX_PARAM_NAME];
	
	int running;
	char *last_event;
	
	pthread_t thread;                         // pthread ID

	float version;
}failover_dev_t;
static failover_dev_t **failover_devices = NULL;

/* baudrates */
typedef struct speed_struct {
    speed_t speed;
    char    *speedstr;
} speed_struct_t;

static speed_struct_t baudrates[] = {
    {B50,         "50"}, {B75,         "75"}, {B110,         "110"},
    {B134,       "134"}, {B150,       "150"}, {B200,         "200"},
    {B300,       "300"}, {B600,       "600"}, {B1200,       "1200"},
    {B1800,     "1800"}, {B2400,     "2400"}, {B4800,       "4800"},
    {B9600,     "9600"}, {B19200,   "19200"}, {B38400,     "38400"},
    {B57600,   "57600"}, {B115200, "115200"}, {0,             NULL}
};

#define FAILOVER_DEV_OPEN 		"open"
#define FAILOVER_DEV_CLOSE 		"close"
#define FAILOVER_DEV_RUNNING	"running"
#define FAILOVER_DEV_TIMER		"time"
	
static int failover_free_dev(void);
static void failover_event_run(struct failover_dev *dev, char *event);

static speed_t parse_baurd(char *speed)
{
	if (!speed)
		return 0;
	
    int i = 0;
    while ((baudrates[i].speedstr) && (strcmp(speed, baudrates[i].speedstr))) 
		i++;
	
    if (baudrates[i].speedstr) {
        return baudrates[i].speed;
    }
	
	return 0;
}

static int parse_parms(tcflag_t *c, char *parms, int flow, int modem)
{
	if (strlen(parms) != 3) {
		ast_log(LOG_WARNING, "Invalid parameter specification \"%s\"!\n", parms);
		return 1;
	}
	
	*c &= ~CSIZE;
	switch(parms[0]) {
	case '5':	*c |= CS5; break;
	case '6':	*c |= CS6; break;
	case '7':	*c |= CS7; break;
	case '8':	*c |= CS8; break;
	default:
		ast_log(LOG_WARNING, "Invalid character size \"%c\": must be 5, 6, 7, or 8!\n",
			parms[0]);
		return 1;
	}
	
	switch(tolower(parms[1])) {
	case 'e':	*c |= PARENB; *c &= ~PARODD; break;
	case 'n':	*c &= ~PARENB;               break;
	case 'o':	*c |= PARENB | PARODD;       break;
	default:
		ast_log(LOG_WARNING, "Invalid parity \"%c\": must be E, N, or O!\n", parms[1]);
		return 1;
	}
	
	switch(parms[2]) {
	case '1':	*c &= ~CSTOPB; break;
	case '2':	*c |= CSTOPB;  break;
	default:
		ast_log(LOG_WARNING, "Invalid stop bit \"%c\": must be 1 or 2!\n", parms[2]);
		return 1;
	}
	
	*c &= ~CRTSCTS;
	if (flow) 
		*c |= CRTSCTS;
	
	*c |= CLOCAL;
	if (modem) 
		*c &= ~CLOCAL;
	
	return 0;
}


static void print_parms(struct termios *ti, char *tty_name)
{
	int baudrate= 0;
	char bits, parity, stops;
	
	/* get speed */
	baudrate = cfgetispeed(ti);
	
	/* get bits */
	switch(ti->c_cflag & CSIZE) {
	case CS5: bits = '5'; break;
	case CS6: bits = '6'; break;
	case CS7: bits = '7'; break;
	case CS8: bits = '8'; break;
	default:
		bits ='?';
	}
	
	/* get parity */
	if (ti->c_cflag & PARENB) {
		parity = ti->c_cflag & PARODD ? 'O' : 'E';
	} else {
		parity = 'N';
	}
	
	/* get stops */
	stops = ti->c_cflag & CSTOPB ? '2' : '1';
	
	/*
	ast_log(LOG_NOTICE, "Connected to %s at %d %c%c%c, modem status %s, %shardware handshake\n",
	tty_name, baudrate, bits, parity, stops, 
	ti->c_cflag & CLOCAL ? "ignored" : "observed",
	ti->c_cflag & CRTSCTS ? "" : "no ");
	*/
}

static int failover_config_tty(struct failover_dev *dev)
{
	tty_struct_t *tty = &(dev->tty);
	char *speed = DEFAULT_SPEED;
	char *parms = DEFAULT_PARMS;
	int flow = 0;
	int modem = 0;
	int qflag = 0;
	struct termios serialti; 
    struct termios tempti;
	int ec = 0;
	int fd;
	speed_t baudrate;
	
	dev->tty.fd = -1;
	
	/* open tty device */
	/*fd = open(tty->tty_name, O_RDWR|O_NONBLOCK|O_NOCTTY);*/
	fd = open(tty->tty_name, O_RDWR|O_NOCTTY);
	if (fd < 0) {
		ast_log(LOG_ERROR, "open %s error!\n", tty->tty_name);
		failover_event_run(dev, FAILOVER_EVENT_ERROR);
		return -1;
	}
	
	/* get tty attr */
	if (tcgetattr(fd, &serialti)) {
		ec = EX_OSERR;
		ast_log(LOG_ERROR, "tcgetattr(%s) error!\n", tty->tty_name);
		goto error;
	}
	
	/* initialize tempti */
	memcpy(&tempti, &serialti, sizeof(tempti));
	
	cfmakeraw(&tempti);
	tempti.c_cc[VMIN] = 1;
	tempti.c_cc[VTIME] = 0;
	
	/* get speed */
	baudrate = parse_baurd(speed);
	if (cfsetispeed(&tempti, baudrate) && cfsetospeed(&tempti, baudrate)) {
		ec = EX_OSERR;
		ast_log(LOG_WARNING,"cfsetspeed(%s) error!\n", tty->tty_name);
		goto error;
	}
	dev->tty.baudrate = baudrate;
	
	/* get parms, flow, modem */
	if (parse_parms(&tempti.c_cflag, parms, flow, modem)) {
		ec = EX_USAGE;
		goto error;
	}
	
	/* set tty device */
	if (tcsetattr(fd, TCSANOW, &tempti)) {
		ec = EX_OSERR;
		ast_log(LOG_WARNING,"cftcsetattr(%s)\n", tty->tty_name);
		goto error;
	}
	
	/* print tty device configure info */
	if (!qflag) {
		if (tcgetattr(fd, &tempti)) {
			close(fd);
			ast_log(LOG_ERROR, "tcgetattr(%s)\n", tty->tty_name);
		}
		print_parms(&tempti, tty->tty_name);
		fflush(stderr);
	}
	cfmakeraw(&tempti);
	
	dev->tty.fd = fd;
    return fd;
error:
	if (fd >= 0) {
		tcsetattr(fd, TCSAFLUSH, &serialti);
		close(fd);
	}
	if (!qflag) 
        ast_log(LOG_ERROR, "Connection closed.\n");
	
	failover_event_run(dev, FAILOVER_EVENT_ERROR);
	
    return -1;
}

static int failover_free_tty(struct failover_dev *dev)
{
	if (!dev) {
		return 0;
	}
	
	int fd = dev->tty.fd;
	if ( fd >= 0) {
		close(fd);
		dev->tty.fd = -1;
	}
	
	return 0;
}

static void failover_event_run(struct failover_dev *dev, char *event)
{
	dev->last_event = event;
	
	if (ast_strlen_zero(dev->event_cmd)) {
		return;
	}
	
	int pid = fork();
	
	switch(pid) {
	case -1: 
		ast_log(LOG_ERROR, "Could not fork event_cmd for failover '%s': %s\n", dev->dev_name, strerror(errno));
		break;
	case 0:
		execlp(dev->event_cmd, dev->event_cmd, dev->dev_name, event, NULL);
		ast_log(LOG_ERROR, "Could not run event_cmd for failover '%s': %s\n", dev->dev_name, strerror(errno));
		
		exit(-1);
	}
}

static int failover_read(struct failover_dev *dev, char *buf)
{
	int i = 0;
	int num;
	char tmpbuf[2];
	
	if (dev->tty.fd) {		
		while((num = read(dev->tty.fd, tmpbuf, 1)) > 0) {
			if (tmpbuf[0] == '\n') {
				tmpbuf[0] = tmpbuf[1] = '\0';
				break;
			} else {
				buf[i++] = tmpbuf[0];
			}
		}
	}

	return i;
}

static void *failover_kick(void *arg)
{
	struct failover_dev *dev = (struct failover_dev *)arg;
	char buf[20];
	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	
	if (dev->tty.fd) {
		usleep(20000);
		
		/* open failover */
		write(dev->tty.fd, FAILOVER_DEV_OPEN, 4);
		failover_read(dev, buf);
		failover_read(dev, buf);
		dev->version = atof(buf + 9);
		usleep(50000);

		/* set watchdog kicking interval */
		snprintf(buf, 9, "time%04d", dev->kick_time_interval);
		buf[8] = '\0';
		write(dev->tty.fd, buf, 8);
		failover_read(dev, buf);
		
		/* kick loop */
		failover_event_run(dev, FAILOVER_EVENT_WDOG_OK);

		while (1) {
			write(dev->tty.fd, FAILOVER_DEV_RUNNING, 7);
			failover_event_run(dev, FAILOVER_EVENT_RUNNING);		
			usleep(dev->kick_time_interval * 1000);
		}
	}else {
		failover_event_run(dev, FAILOVER_EVENT_WDOG_FAILED);
	}

	return NULL;
}

static void failover_kick_start(struct failover_dev *dev)
{
	int ret = -1;
	
	if (dev->running)
		return;
	dev->running = 1;
	
	/* configure tty device */
	ret = failover_config_tty(dev);
	
    /* create pthread */
	if (ret >= 0)
		ast_pthread_create(&dev->thread, NULL, failover_kick, dev);
}

static void failover_kick_stop(struct failover_dev *dev)
{	
	char buf[20];
	
	if (!dev || !dev->running)
		return;
	dev->running = 0;
	
	
    /* cancel pthread  */
    if (dev->tty.fd >= 0) {
		pthread_cancel(dev->thread);
		pthread_join(dev->thread, NULL);
    }
	
	/* close failover */
	write(dev->tty.fd, FAILOVER_DEV_CLOSE, 5);	
	failover_event_run(dev, FAILOVER_EVENT_STOPPED);

	failover_read(dev, buf);
	usleep(20000);

	/* free tty device */
	failover_free_tty(dev);
}

static int failover_load_config(void)
{
    struct ast_config *cfg ;
    char *cat = NULL;
    failover_dev_t *dev = NULL;
	int ret = -1;
	
    failover_free_dev();    
	
    /* load config */
#ifdef AST_VERSION_1_6
	struct ast_flags config_flags = { 0 };
	cfg = ast_config_load(FAILOVER_CONFIG_FILE, config_flags);
#else
    cfg = ast_config_load(FAILOVER_CONFIG_FILE);
#endif
    if (cfg == NULL) {
        ast_log(LOG_ERROR, "Could not load failover config file "FAILOVER_CONFIG_FILE"!\n");
        return -1;
    }
	
    /* get config parameters */
    while ((cat = ast_category_browse(cfg, cat))) {
        /* get device name */
		const char *failover_dev_name = cat;
		
        /* get kick_time_interval */
		const char *tty_name = ast_variable_retrieve(cfg, cat, "device");
		
        /* get kick_time_interval */
		const char *kick_time_interval = ast_variable_retrieve(cfg, cat, "kick_time_interval");
		
        /* get event_cmd */
		const char *event_cmd = ast_variable_retrieve(cfg, cat, "event_cmd");
		
        /* get autorun */
		const char *autorun = ast_variable_retrieve(cfg, cat, "autorun");
		
		if (failover_dev_name && tty_name) {
			/* malloc a dev*/
			dev = (failover_dev_t *)ast_malloc(sizeof(failover_dev_t));
			if (!dev) {
				ast_log(LOG_ERROR, "malloc memory error!\n");
				return -1;
			}
			
			/* initialize dev */
			memset(dev, 0, sizeof(failover_dev_t));
			
			/* set failover dev_name */
			ast_copy_string(dev->dev_name, failover_dev_name, MAX_DEVICE_NAME - 1);
			
			/* set failover tty_name */
			ast_copy_string(dev->tty.tty_name, tty_name, MAX_DEVICE_NAME - 1);
			
			/* set watchdog interval */
			int interval;			
			if (ast_strlen_zero(kick_time_interval)) {
				interval = 4000;
			} else {
				interval = atoi(kick_time_interval);
				if((interval < 100) || (interval > 9000))
					interval = 4000;
			}
			dev->kick_time_interval = interval;
			
			/* set event_cmd */
			if (ast_strlen_zero(event_cmd)) {
				ast_copy_string(dev->event_cmd, "logger", MAX_CMD_NAME - 1);
			} else if (strcasecmp(event_cmd, "no")){
				ast_copy_string(dev->event_cmd, event_cmd, MAX_CMD_NAME - 1);
			}
			
			dev->last_event = FAILOVER_EVENT_STOPPED;
			dev->running    = 0;
			
			/* set autorun */
			if (ast_strlen_zero(autorun)) {
				ast_copy_string(dev->autorun, "yes", MAX_PARAM_NAME - 1);
			} else {
				ast_copy_string(dev->autorun, autorun, MAX_PARAM_NAME - 1);
			}
			
			failover_dev_num ++;
			failover_devices = (failover_dev_t **)ast_realloc(failover_devices, sizeof(failover_dev_t *) * failover_dev_num);
			failover_devices[failover_dev_num - 1] = dev;
			
			if (!strcasecmp(dev->autorun, "yes")) {
				failover_kick_start(dev);
			}
			
			ret = 0;
		}
    }
	
	if (ret < 0) {
		ast_log(LOG_ERROR, "load config file %s error\n", FAILOVER_CONFIG_FILE);
	}
	
	/* destroy config loaded*/
    ast_config_destroy(cfg);
	
    return 0;
}

static int failover_free_dev(void)
{
	int i;
	
    if((0 == failover_dev_num) && (NULL == failover_devices))
        return 0;
	
	for (i = 0; i < failover_dev_num; i ++) {
		ast_free(failover_devices[i]);
	}
	
	ast_free(failover_devices);
    failover_devices = NULL;
	
    failover_dev_num = 0;
	
	return 0;
}

static void failover_show(const int fd, struct failover_dev *dev, int i) 
{  
	if (dev) {
		if (dev->version) {
        ast_cli(fd, " * DEVICE#%d: %s %s VERSION:%.1f\n",
			i + 1,
			dev->dev_name,
			dev->tty.tty_name,
			dev->version);
		} else {
	        ast_cli(fd, " * DEVICE#%d: %s %s\n",
				i + 1,
				dev->dev_name,
				dev->tty.tty_name);
		}
		
        ast_cli(fd, "   INTERVAL:%d EVENT:%s AUTORUN:%s STATE:%s\n",
			dev->kick_time_interval,
			dev->event_cmd,
			dev->autorun,
			dev->last_event);	
	}
}

static void failover_start(const int fd, struct failover_dev *dev) 
{	
	/* kick start */
	ast_cli(fd, "'%s' opened!\n", dev->dev_name);
	failover_kick_start(dev);
}

static void failover_stop(const int fd, struct failover_dev *dev) 
{
	/* kick stop */
	ast_cli(fd, "'%s' closed!\n", dev->dev_name);
	failover_kick_stop(dev);
}

#ifdef AST_VERSION_1_6
static char *failover_cli_handle_start(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char *ret = CLI_FAILURE;
	int result;
    int i;
    
	switch (cmd) {
	case CLI_INIT:
		e->command = "failover start";
		e->usage =
            "Usage: failover start [device number]\n"
            "       Example:\n"    
            "       failover start\n"
            "       failover start 1";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (0 == failover_dev_num) {
		ast_log(LOG_ERROR, "load config file %s error\n", FAILOVER_CONFIG_FILE);
		return ret;
	}
	
    if (2 == a->argc) {
		for (i = 0; i < failover_dev_num; i ++) {
			failover_start(a->fd, failover_devices[i]);
		}
		ret = RESULT_SUCCESS;
    } else if (3 == a->argc) {
		result = atoi(a->argv[2]);
		if (result >= 1 && result <= failover_dev_num) {
			failover_start(a->fd, failover_devices[result - 1]);
    		ret = CLI_SUCCESS;          
		} else {
			ret = CLI_SHOWUSAGE;
		}
	} else {
		ret = CLI_SHOWUSAGE;
	}		
	
	return ret;
	
}

static char *failover_cli_handle_stop(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char *ret = CLI_FAILURE;
	int result;
    int i;
    
	switch (cmd) {
	case CLI_INIT:
		e->command = "failover stop";
		e->usage =
            "Usage: failover stop [device number]\n"
            "       Example:\n"    
            "       failover stop\n"
            "       failover stop 1";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

    if (2 == a->argc) {
		for (i = 0; i < failover_dev_num; i ++) {
			failover_stop(a->fd, failover_devices[i]);
		}
		ret = CLI_SUCCESS;
    } else if (3 == a->argc) {
		result = atoi(a->argv[2]);
		if (result >= 1 && result <= failover_dev_num) {
			failover_stop(a->fd, failover_devices[result - 1]);
    		ret = CLI_SUCCESS;            
		} else {
			ret = CLI_SHOWUSAGE;
		}
	} else {
		ret = CLI_SHOWUSAGE;
	}		
	
	return ret;
}

static char *failover_cli_handle_show(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char *ret = CLI_FAILURE;
    int i;
    
	switch (cmd) {
	case CLI_INIT:
		e->command = "failover show";
		e->usage =
            "Usage: failover show\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

    if (2 == a->argc) {
		for (i = 0; i < failover_dev_num; i ++) {
			failover_show(a->fd, failover_devices[i], i);
		}
		ret = CLI_SUCCESS;
    } else {
		ret = CLI_SHOWUSAGE;
	}		
	
	return ret;
}

static char *failover_cli_handle_reload(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char *ret = CLI_FAILURE;
    int i;
    
	switch (cmd) {
	case CLI_INIT:
		e->command = "failover reload";
		e->usage =
            "Usage: failover reload\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (2 == a->argc) {
		for (i = 0; i < failover_dev_num; i ++) {
			if(strcasecmp(failover_devices[i]->last_event, FAILOVER_EVENT_STOPPED)) {
		    	ast_log(LOG_ERROR, "You must stop the failover device first!\n");
				return ret;
			}
		}

		if(failover_load_config() < 0) {
		    ast_log(LOG_ERROR, "Could not initialize failover module!");
		    return ret;
		}
		ret = CLI_SUCCESS;
		ast_cli(a->fd, "Reload config file successfully!\n");
	} else {
		ret = CLI_SHOWUSAGE;
	}	
	return ret;
}
#else
static int failover_cli_handle_start(int fd, int argc, char *argv[]) 
{
	int ret = RESULT_FAILURE;
	int result;
    int i;
	
	if (0 == failover_dev_num) {
		ast_log(LOG_ERROR, "load config file %s error\n", FAILOVER_CONFIG_FILE);
		return ret;
	}
	
    if (2 == argc) {
		for (i = 0; i < failover_dev_num; i ++) {
			failover_start(fd, failover_devices[i]);
		}
		ret = RESULT_SUCCESS;
    } else if (3 == argc) {
		result = atoi(argv[2]);
		if (result >= 1 && result <= failover_dev_num) {
			failover_start(fd, failover_devices[result - 1]);
    		ret = RESULT_SUCCESS;          
		} else {
			ret = RESULT_SHOWUSAGE;
		}
	} else {
		ret = RESULT_SHOWUSAGE;
	}		
	
	return ret;
}

static int failover_cli_handle_stop(int fd, int argc, char *argv[]) 
{
	int ret = RESULT_FAILURE;
	int result;
    int i;
	
    if (2 == argc) {
		for (i = 0; i < failover_dev_num; i ++) {
			failover_stop(fd, failover_devices[i]);
		}
		ret = RESULT_SUCCESS;
    } else if (3 == argc) {
		result = atoi(argv[2]);
		if (result >= 1 && result <= failover_dev_num) {
			failover_stop(fd, failover_devices[result - 1]);
    		ret = RESULT_SUCCESS;            
		} else {
			ret = RESULT_SHOWUSAGE;
		}
	} else {
		ret = RESULT_SHOWUSAGE;
	}		
	
	return ret;
}

static int failover_cli_handle_show(int fd, int argc, char *argv[]) 
{
	int ret = RESULT_FAILURE;
    int i;
	
    if (2 == argc) {
		for (i = 0; i < failover_dev_num; i ++) {
			failover_show(fd, failover_devices[i], i);
		}
		ret = RESULT_SUCCESS;
    } else {
		ret = RESULT_SHOWUSAGE;
	}		
	
	return ret;
}

static int failover_cli_handle_reload(int fd, int argc, char *argv[]) 
{
	int ret = RESULT_FAILURE;
	int i;
	
	if (2 == argc) {
		for (i = 0; i < failover_dev_num; i ++) {
			if(strcasecmp(failover_devices[i]->last_event, FAILOVER_EVENT_STOPPED)) {
		    	ast_log(LOG_ERROR, "You must stop the failover device first!\n");
				return ret;
			}
		}

		if(failover_load_config() < 0) {
		    ast_log(LOG_ERROR, "Could not initialize failover module!");
		    return ret;
		}
		ret = RESULT_SUCCESS;
		ast_cli(fd, "Reload config file successfully!\n");
	} else {
		ret = RESULT_SHOWUSAGE;
	}	
    
	return ret;
}
#endif

#ifdef AST_VERSION_1_6
static struct ast_cli_entry cli_failover[] = {
	AST_CLI_DEFINE(failover_cli_handle_start, "Starts failover device"),
	AST_CLI_DEFINE(failover_cli_handle_stop, "Stops failover device"),
	AST_CLI_DEFINE(failover_cli_handle_show, "Shows failover config from failover.conf"),
	AST_CLI_DEFINE(failover_cli_handle_reload, "Reloads failover config from failover.conf"),
};
#else
static char usage_failover_start[] =
"Usage: failover start [device number]\n"
"       Example:\n"
"           failover start\n"
"           failover start 1\n"
;

static char usage_failover_stop[] =
"Usage: failover stop [device number]\n"
"       Example:\n"
"           failover stop\n"
"           failover stop 1\n"
;

static char usage_failover_show[] =
"Usage: failover show\n"
;

static char usage_failover_reload[] =
"Usage: failover reload\n"
;

static struct ast_cli_entry cli_failover[] = {
	{ 
		{ "failover", "start" , NULL },
		failover_cli_handle_start, 
		"Starts failover device",
		usage_failover_start
	},
	{ 
		{ "failover", "stop" , NULL },
		failover_cli_handle_stop, 
		"Stops failover device",
		usage_failover_stop
	},
	{ 
		{ "failover", "show" , NULL },
		failover_cli_handle_show, 
		"Shows failover config from failover.conf",
		usage_failover_show
	},
	{ 
		{ "failover", "reload" , NULL },
		failover_cli_handle_reload, 
		"Reloads failover config from failover.conf",
		usage_failover_reload
	},
};
#endif

__static__ int load_module(void)
{
    int ret;
	
    ast_cli_register_multiple(cli_failover, sizeof(cli_failover) / sizeof(struct ast_cli_entry));
	
    ret = failover_load_config();
    if(ret < 0) {
        ast_log(LOG_ERROR, "Could not initialize failover module!");
        return ret;
    }
    
    return 0;
}

__static__ int unload_module(void)
{
	int i;
	
	for (i = 0; i < failover_dev_num; i ++) {
		failover_kick_stop(failover_devices[i]);		
	}
	
    failover_free_dev();
	
    ast_cli_unregister_multiple(cli_failover, sizeof(cli_failover) / sizeof(struct ast_cli_entry));
	
    return 0;
}


__static__ int reload (void) 
{
	return -1;
}

#ifdef AST_VERSION_1_2
char *description(void)
{
	return "OpenVox Failover Resource";
}

int usecount(void)
{
	return 1;
}
char *key()
{
	return ASTERISK_GPL_KEY;
}
#else
AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS, "OpenVox Failover Device",
				.load = load_module,
				.unload = unload_module,
				.reload = reload,
				);
#endif



